# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class TooManyResult(Exception):
    pass


def search_one(data, **attrs):
    """Returns on element of data array that have the same attrs that in **attrs

    Data should be in form array of dictionary:

    data = [
        {
            attr1: attr1_value,
            attr2: attr2_value,
            ...
        },
        ...
    ]

    **attrs is a values to filter something

    search_one return one element from data if all **attrs equal to attrs
    from dict. It doesn't metter if len(**attrs) < len(element)

    search_one returns None if no result found

    search_one raises TooManyResult if more than one element in result
    """
    result = None
    for i in search(data, **attrs):
        if result is None:
            result = i
        else:
            # if we have one element in search generator than
            # we should not be here
            raise TooManyResult
    return result


def search(data, **attrs):
    """Return subset of data array with elements that have
    the same attrs that in **attrs

    Data should be in form array of dictionary:

    data = [
        {
            attr1: attr1_value,
            attr2: attr2_value,
            ...
        },
        ...
    ]

    **attrs is a values to filter something

    search return array of elements elements from data
    if all **attrs are equal to attrs of element.
    It doesn't metter if len(**attrs) < len(element)

    search returns None if no result found
    """
    return filter(lambda x: attrs.items() <= x.items(), data)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=R0901, R0904
import logging
from contextlib import contextmanager

from coreapi.exceptions import ErrorMessage
from version_utils import rpm

from adcm_client.base import (ActionHasIssues, ADCMApiError, BaseAPIListObject,
                              BaseAPIObject, ObjectNotFound, TooManyArguments,
                              strip_none_keys)
from adcm_client.util import stream
from adcm_client.wrappers.api import ADCMApiWrapper

# If we are running the client from tests with Allure we expected that code
# to trace steps in Allure UI.
# But in case of running client outside of testing Allure is useless in virtualenv.
# So that code should be flexible enought to work with Allure or without.
ALLURE = True
try:
    import allure
except ImportError:
    ALLURE = False

# Init logger
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


# That is trick which is allmost the same that in _allure.py::StepContext
# We have a function that can be used as contextmanager and decorator
# in same time.
@contextmanager
def dummy_context(text):
    yield text


def allure_step(text):
    if ALLURE:
        return allure.step(text)
    return dummy_context(text)


class NoCredentionsProvided(Exception):
    """There is no user/password provided. It was not passsed as init parameters
    during ADCMClient initialization and was not passed as a parameter to auth()
    function
    """


##################################################
#                 B U N D L E S
##################################################
class IncorrectPrototypeType(Exception):
    pass


class Bundle(BaseAPIObject):
    IDNAME = "bundle_id"
    PATH = ["stack", "bundle"]
    FILTERS = ["name", "version"]

    id = None
    bundle_id = None
    name = None
    description = None
    version = None
    edition = None

    def provider_prototype(self) -> "ProviderPrototype":
        return self._child_obj(ProviderPrototype)

    def provider_create(self, name, description=None) -> "Provider":
        try:
            prototype = self.provider_prototype()
        except ObjectNotFound:
            raise IncorrectPrototypeType
        return prototype.provider_create(name, description)

    def provider_list(self, paging=None, **args) -> "ProviderList":
        try:
            prototype = self.provider_prototype()
        except ObjectNotFound:
            raise IncorrectPrototypeType
        return prototype.provider_list(paging=paging, **args)

    def provider(self, **args) -> "Provider":
        try:
            prototype = self.provider_prototype()
        except ObjectNotFound:
            raise IncorrectPrototypeType
        return prototype.provider(**args)

    def service_prototype(self, **args) -> "ServicePrototype":
        return self._child_obj(ServicePrototype, **args)

    def cluster_prototype(self) -> "ClusterPrototype":
        return self._child_obj(ClusterPrototype)

    def cluster_create(self, name, description=None) -> "Cluster":
        try:
            prototype = self.cluster_prototype()
        except ObjectNotFound:
            raise IncorrectPrototypeType
        return prototype.cluster_create(name, description)

    def cluster_list(self, paging=None, **args) -> "ClusterList":
        try:
            prototype = self.cluster_prototype()
        except ObjectNotFound:
            raise IncorrectPrototypeType
        return prototype.cluster_list(paging=paging, **args)

    def cluster(self, **args) -> "Cluster":
        try:
            prototype = self.cluster_prototype()
        except ObjectNotFound:
            raise IncorrectPrototypeType
        return prototype.cluster(**args)

    def license(self):
        return self._subcall("license", "read")

    def license_accept(self):
        self._subcall("license", "accept", "update")


class BundleList(BaseAPIListObject):
    _ENTRY_CLASS = Bundle


##################################################
#              P R O T O T Y P E
##################################################
class Prototype(BaseAPIObject):
    PATH = ["stack", "prototype"]
    IDNAME = "prototype_id"
    FILTERS = ["name", "bundle_id"]

    id = None
    prototype_id = None
    name = None
    type = None
    description = None
    version = None
    bundle_id = None
    config = None
    actions = None
    url = None

    def bundle(self) -> "Bundle":
        return self._parent_obj(Bundle)


class PrototypeList(BaseAPIListObject):
    _ENTRY_CLASS = Prototype


class ClusterPrototype(Prototype):
    PATH = ["stack", "cluster"]
    FILTERS = ["name", "bundle_id"]

    def cluster_create(self, name, description=None) -> "Cluster":
        if self.type != 'cluster':
            raise IncorrectPrototypeType
        return new_cluster(
            self._api,
            prototype_id=self.prototype_id,
            name=name,
            description=description
        )

    def cluster_list(self, paging=None, **args) -> "ClusterList":
        return self._child_obj(ClusterList, paging=paging, **args)

    def cluster(self, **args) -> "Cluster":
        return self._child_obj(Cluster, **args)


class ClusterPrototypeList(BaseAPIListObject):
    _ENTRY_CLASS = ClusterPrototype


class ServicePrototype(Prototype):
    PATH = ["stack", "service"]
    FILTERS = ["name", "bundle_id"]

    shared = None
    display_name = None
    required = None
    components = None
    exports = None
    imports = None
    monitoring = None

    def service_list(self, paging=None, **args) -> "ServiceList":
        return self._child_obj(ServiceList, paging=paging, **args)

    def service(self, **args) -> "Service":
        return self._child_obj(Service)


class ServicePrototypeList(BaseAPIListObject):
    _ENTRY_CLASS = ServicePrototype


class ProviderPrototype(Prototype):
    PATH = ["stack", "provider"]
    FILTERS = ["name", "bundle_id"]

    display_name = None
    required = None
    upgrade = None

    def provider_create(self, name, description=None) -> "Provider":
        if self.type != 'provider':
            raise IncorrectPrototypeType
        return new_provider(
            self._api,
            prototype_id=self.prototype_id,
            name=name,
            description=description
        )

    def provider_list(self, paging=None, **args) -> "ProviderList":
        return self._child_obj(ProviderList, paging=paging, **args)

    def provider(self, **args) -> "Provider":
        return self._child_obj(Provider, **args)


class ProviderPrototypeList(BaseAPIListObject):
    _ENTRY_CLASS = ProviderPrototype


class HostPrototype(Prototype):
    PATH = ["stack", "host"]
    FILTERS = ["name", "bundle_id"]

    display_name = None
    required = None
    monitoring = None

    def host_list(self, paging=None, **args) -> "HostList":
        return self._child_obj(HostList, paging=paging, **args)

    def host(self, **args) -> "Host":
        return self._child_obj(Host, **args)


class HostPrototypeList(BaseAPIListObject):
    _ENTRY_CLASS = HostPrototype


##################################################
#           B A S E  O B J E C T
##################################################
class _BaseObject(BaseAPIObject):
    id = None
    url = None
    state = None
    prototype_id = None
    issue = None
    button = None

    def prototype(self):
        raise NotImplementedError

    def action(self, **args) -> "Action":
        return self._subobject(Action, **args)

    def action_list(self, paging=None, **args) -> "ActionList":
        return self._subobject(ActionList, paging=paging, **args)

    def action_run(self, **args) -> "Task":
        action = self.action(**args)
        return action.run()

    def config(self, full=False):
        history_entry = self._subcall("config", "current", "list")
        if full:
            return history_entry
        return history_entry['config']

    @allure_step("Save config")
    def config_set(self, data):
        if "config" in data and "attr" in data:
            # We are in a new mode with full_info == True
            if data["attr"] is None:
                data["attr"] = []
            history_entry = self._subcall(
                'config', 'history', 'create',
                config=data["config"],
                attr=data["attr"],
            )
            return history_entry
        history_entry = self._subcall('config', 'history', 'create', config=data)
        return history_entry['config']

    @allure_step("Save config")
    def config_set_diff(self, data):
        config = self.config()
        for gk, gv in data.items():
            if isinstance(gv, dict):
                for k, v in gv.items():
                    config[gk][k] = v
            else:
                config[gk] = gv
        self.config_set(config)

    def config_prototype(self):
        return self.prototype().config


##################################################
#              P R O V I D E R
##################################################
class Provider(_BaseObject):
    IDNAME = "provider_id"
    PATH = ["provider"]
    FILTERS = ["name", "prototype_id"]
    provider_id = None
    name = None
    description = None
    bundle_id = None

    def bundle(self) -> "Bundle":
        return self._parent_obj(Bundle)

    def host_create(self, fqdn) -> "Host":
        return new_host(self._api, **self._merge(fqdn=fqdn))

    def host_list(self, paging=None, **args) -> "HostList":
        return self._child_obj(HostList, paging=paging, **args)

    def host(self, **args) -> "Host":
        return self._child_obj(Host, **args)

    def prototype(self) -> "ProviderPrototype":
        return self._parent_obj(ProviderPrototype)

    def upgrade(self, **args) -> "Upgrade":
        return self._subobject(Upgrade, **args)

    def upgrade_list(self, paging=None, **args) -> "UpgradeList":
        return self._subobject(UpgradeList, paging=paging, **args)


class ProviderList(BaseAPIListObject):
    _ENTRY_CLASS = Provider


@allure_step('Create provider {name}')
def new_provider(api, **args) -> "Provider":
    p = api.objects.provider.create(**strip_none_keys(args))
    return Provider(api, provider_id=p['id'])


##################################################
#              C L U S T E R
##################################################
class Cluster(_BaseObject):
    IDNAME = "cluster_id"
    PATH = ["cluster"]
    FILTERS = ["name", "prototype_id"]
    cluster_id = None
    name = None
    description = None
    bundle_id = None
    serviceprototype = None
    status = None

    def prototype(self) -> "ClusterPrototype":
        return self._parent_obj(ClusterPrototype)

    def bind(self, target):
        if isinstance(target, Cluster):
            self._subcall("bind", "create", export_cluster_id=target.cluster_id)
        elif isinstance(target, Service):
            self._subcall("bind", "create", export_cluster_id=target.cluster_id,
                          export_service_id=target.service_id)
        else:
            raise NotImplementedError

    def bind_list(self, paging=None):
        return self._subcall("bind", "list")

    def bundle(self) -> "Bundle":
        proto = self.prototype()
        return proto.bundle()

    def button(self):
        raise NotImplementedError

    def host(self, **args) -> "Host":
        return self._child_obj(Host, **args)

    def host_list(self, paging=None, **args) -> "HostList":
        return self._child_obj(HostList, paging=paging, **args)

    def host_add(self, host: "Host") -> "Host":
        with allure_step("Add host {} to cluster {}".format(host.fqdn, self.name)):
            data = self._subcall("host", "create", host_id=host.id)
            return Host(self._api, id=data['id'])

    def host_delete(self, host: "Host"):
        with allure_step("Remove host {} from cluster {}".format(host.fqdn, self.name)):
            self._subcall("host", "delete", host_id=host.id)

    def service(self, **args) -> "Service":
        return self._subobject(Service, **args)

    def service_list(self, paging=None, **args) -> "ServiceList":
        return self._subobject(ServiceList, paging=paging, **args)

    def service_add(self, **args) -> "Service":
        proto = self.bundle().service_prototype(**args)
        with allure_step("Add service {} to cluster {}".format(proto.name, self.name)):
            data = self._subcall("service", "create", prototype_id=proto.id)
            return self._subobject(Service, service_id=data['id'])

#    @allure_step("Remove service to cluster")
#    def service_delete(self, service):
#        self._subcall("service", "delete", service_id=service.id)

    def hostcomponent(self):
        return self._subcall("hostcomponent", "list")

    @allure_step("Save hostcomponents map")
    def hostcomponent_set(self, *hostcomponents):
        hc = []
        for i in hostcomponents:
            h, c = i
            hc.append({
                'host_id': h.id,
                'service_id': c.service_id,
                'component_id': c.id
            })
        return self._subcall("hostcomponent", "create", hc=hc)

    def status_url(self):
        return self._subcall("status", "list")

    def imports(self):
        raise NotImplementedError

    def upgrade(self, **args) -> "Upgrade":
        return self._subobject(Upgrade, **args)

    def upgrade_list(self, paging=None, **args) -> "UpgradeList":
        return self._subobject(UpgradeList, paging=paging, **args)


class ClusterList(BaseAPIListObject):
    _ENTRY_CLASS = Cluster


@allure_step('Create cluster {name}')
def new_cluster(api: ADCMApiWrapper, **args) -> "Cluster":
    c = api.objects.cluster.create(**strip_none_keys(args))
    return Cluster(api, cluster_id=c['id'])


##################################################
#          U P G R A D E
##################################################
class Upgrade(BaseAPIObject):
    IDNAME = "upgrade_id"
    PATH = None
    SUBPATH = ["upgrade"]

    id = None
    upgrade_id = None
    url = None
    name = None
    description = None
    min_version = None
    max_version = None
    min_strict = None
    max_strict = None
    upgradable = None
    state_available = None
    state_on_success = None
    from_edition = None

    def do(self, **args):
        with allure_step("Do upgrade {}".format(self.name)):
            self._subcall("do", "create", **args)


class UpgradeList(BaseAPIListObject):
    SUBPATH = ["upgrade"]
    _ENTRY_CLASS = Upgrade


##################################################
#           S E R V I C E S
##################################################
class Service(_BaseObject):
    IDNAME = "service_id"
    PATH = None
    SUBPATH = ["service"]

    id = None
    service_id = None
    bundle_id = None
    name = None
    description = None
    display_name = None
    cluster_id = None
    status = None
    button = None
    monitoring = None

    def bind(self, target):
        if isinstance(target, Cluster):
            self._subcall("bind", "create", export_cluster_id=target.cluster_id)
        elif isinstance(target, Service):
            self._subcall("bind", "create", export_cluster_id=target.cluster_id,
                          export_service_id=target.service_id)
        else:
            raise NotImplementedError

    def prototype(self) -> "ServicePrototype":
        return ServicePrototype(self._api, id=self.prototype_id)

    def imports(self):
        raise NotImplementedError

    def bind_list(self, paging=None):
        return self._subcall("bind", "list")

    def component(self, **args) -> "Component":
        return self._subobject(Component, **args)

    def component_list(self, paging=None, **args) -> "ComponentList":
        return self._subobject(ComponentList, paging=paging, **args)


class ServiceList(BaseAPIListObject):
    SUBPATH = ["service"]
    _ENTRY_CLASS = Service


##################################################
#           C O M P O N E N T S
##################################################
class Component(BaseAPIObject):
    IDNAME = "component_id"
    SUBPATH = ["component"]

    id = None
    component_id = None
    name = None
    display_name = None
    description = None
    constraint = None
    params = None
    prototype_id = None

    @property
    def service_id(self):
        return self._endpoint.path_args["service_id"]


class ComponentList(BaseAPIListObject):
    SUBPATH = ["component"]
    _ENTRY_CLASS = Component


##################################################
#              H O S T
##################################################
class Host(_BaseObject):
    IDNAME = "host_id"
    PATH = ["host"]
    FILTERS = ["fqdn", "prototype_id", "provider_id", "cluster_id"]

    id = None
    host_id = None
    fqdn = None
    provider_id = None
    cluster_id = None
    description = None
    bundle_id = None
    status = None

    def provider(self) -> "Provider":
        return self._parent_obj(Provider)

    def cluster(self) -> "Cluster":
        return self._parent_obj(Cluster)

    def bundle(self) -> "Bundle":
        return self._parent_obj(Bundle)

    def prototype(self) -> "HostPrototype":
        return self._parent_obj(HostPrototype)


class HostList(BaseAPIListObject):
    _ENTRY_CLASS = Host


@allure_step('Create host {fqdn}')
def new_host(api, **args) -> "Host":
    h = api.objects.provider.host.create(**args)
    return Host(api, host_id=h['id'])


##################################################
#              A C T I O N
##################################################
class Action(BaseAPIObject):
    IDNAME = "action_id"
    PATH = None
    SUBPATH = ["action"]
    FILTERS = ["name"]

    action_id = None
    button = None
    id = None
    name = None
    display_name = None
    description = None
    params = None
    prototype_id = None
    required_hostcomponentmap = None
    hostcomponentmap = None
    script = None
    script_type = None
    state_available = None
    state_on_fail = None
    state_on_success = None
    type = None
    url = None
    subs = None
    config = None

    def _get_config(self):
        config = {}
        for item in self.config['config']:
            config[item['name']] = item['value']
        return config

    def log_files(self):
        raise NotImplementedError

    def task(self, **args) -> "Task":
        return self._child_obj(Task, **args)

    def task_list(self, **args) -> "TaskList":
        return self._child_obj(TaskList, **args)

    def run(self, **args) -> "Task":
        with allure_step("Run action {}".format(self.name)):

            if 'config' in args and 'config_diff' in args:
                raise TypeError("only one argument is expected 'config' or 'config_diff'")

            if 'config' not in args:
                args['config'] = self._get_config()

                if 'config_diff' in args:
                    config_diff = args.pop('config_diff')
                    for key, value in config_diff.items():
                        args['config'][key] = value

            try:
                data = self._subcall("run", "create", **args)
            except ErrorMessage as error:
                if (error.title == '409 Conflict'
                        and 'has issues' in getattr(error, '_data', {}).get('desc', '')):
                    raise ActionHasIssues
                raise error
            return Task(self._api, task_id=data["id"])


class ActionList(BaseAPIListObject):
    SUBPATH = ["action"]
    _ENTRY_CLASS = Action


##################################################
#              T A S K
##################################################
class TaskFailed(Exception):
    pass


class Task(BaseAPIObject):
    IDNAME = "task_id"
    PATH = ["task"]
    FILTERS = ['action_id', 'pid', 'status', 'start_date', 'finish_date']
    _END_STATUSES = ["failed", "success"]
    action = None
    action_id = None
    config = None
    hostcomponentmap = None
    task_id = None
    id = None
    jobs = None
    pid = None
    selector = None
    status = None
    url = None

    def job(self, **args) -> "Job":
        return Job(self._api, path_args=dict(task_id=self.id), **args)

    def job_list(self, paging=None, **args) -> "JobList":
        return JobList(self._api, paging=paging, path_args=dict(task_id=self.id), **args)

    @allure_step("Wait for task end")
    def wait(self, timeout=None):
        return self.wait_for_attr("status",
                                  self._END_STATUSES,
                                  timeout=timeout)

    @allure_step("Wait for task to success.")
    def try_wait(self, timeout=None):
        status = self.wait(timeout=timeout)

        if status == "failed":
            for job in self.job_list(status="failed"):
                for file in job.log_files:
                    logger.error(self._api.client.get(file["url"])["content"])
            raise TaskFailed

        return status


class TaskList(BaseAPIListObject):
    _ENTRY_CLASS = Task


##################################################
#              J O B
##################################################
class Job(BaseAPIObject):
    IDNAME = "job_id"
    PATH = ["job"]
    FILTERS = ['action_id', 'task_id', 'pid', 'status', 'start_date', 'finish_date']
    _END_STATUSES = ["failed", "success"]
    _WAIT_INTERVAL = .2
    id = None
    job_id = None
    pid = None
    status = None
    url = None
    log_files = None
    task_id = None

    def __init__(self, api: ADCMApiWrapper, path=None, path_args=None, **args):
        super().__init__(api, path, **args)

    def wait(self, timeout=None):
        return self.wait_for_attr("status",
                                  self._END_STATUSES,
                                  timeout=timeout)


class JobList(BaseAPIListObject):
    _ENTRY_CLASS = Job


##################################################
#              A D C M
##################################################
class ADCM(BaseAPIObject):
    IDNAME = "adcm_id"
    PATH = ["adcm"]
    id = None
    name = None
    prototype_id = None
    url = None
    prototype_version = None
    bundle_id = None

    # TODO: Remove that function when it become first class object
    def config(self, full=False):
        history_entry = self._subcall("config", "current", "list")
        if full:
            return history_entry
        return history_entry['config']

    @allure_step("Save config")
    def config_set(self, data):
        if "config" in data and "attr" in data:
            # We are in a new mode with full_info == True
            if data["attr"] is None:
                data["attr"] = []
            history_entry = self._subcall(
                'config', 'history', 'create',
                config=data["config"],
                attr=data["attr"],
            )
            return history_entry
        history_entry = self._subcall('config', 'history', 'create', config=data)
        return history_entry['config']

    @allure_step("Save config")
    def config_set_diff(self, data):
        config = self.config()
        for gk, gv in data.items():
            if isinstance(gv, dict):
                for k, v in gv.items():
                    config[gk][k] = v
            else:
                config[gk] = gv
        self.config_set(config)

    def config_prototype(self):
        return self.prototype().config


##################################################
#              C L I E N T
##################################################
class ADCMClient:
    _MIN_VERSION = "2019.10.01.00"

    def __init__(self, api=None, url=None, user=None, password=None):
        if api is not None:
            self._api = api
            self.url = api.url
            if self.api_token() is None:
                self.auth(user, password)
        else:
            self.url = url
            self._api = ADCMApiWrapper(self.url)
            self.auth(user, password)
        if self.api_token() is not None:
            self.guess_adcm_url()

    def auth(self, user=None, password=None):
        if user is None or password is None:
            raise NoCredentionsProvided
        self._api.auth(user, password)
        if self.api_token() is None:
            raise ADCMApiError("Incorrect user/password. Unable to auth.")
        self._check_min_version()

    def _check_min_version(self):
        try:
            version = self._api.objects.info.list()['adcm_version']
        except (KeyError, AttributeError):
            version = "0"
        if rpm.compare_versions(self._MIN_VERSION, version) > 0:
            raise ADCMApiError("That client supports ADCM versions >= {}".format(self._MIN_VERSION))

    def api_token(self):
        return self._api.api_token

    def adcm(self) -> ADCM:
        return ADCM(self._api)

    def guess_adcm_url(self):
        config = self.adcm().config()
        if config['global']['adcm_url'] is None:
            self.adcm().config_set_diff({"global": {"adcm_url": self.url}})

    def bundle(self, **args) -> Bundle:
        return Bundle(self._api, **args)

    def bundle_list(self, paging=None, **args) -> BundleList:
        return BundleList(self._api, paging=paging, **args)

    def cluster(self, **args) -> Cluster:
        return Cluster(self._api, **args)

    def cluster_list(self, paging=None, **args) -> ClusterList:
        return ClusterList(self._api, paging=paging, **args)

    def cluster_prototype(self, **args) -> ClusterPrototype:
        return ClusterPrototype(self._api, **args)

    def cluster_prototype_list(self, paging=None, **args) -> ClusterPrototypeList:
        return ClusterPrototypeList(self._api, paging=paging, **args)

    def host(self, **args) -> Host:
        return Host(self._api, **args)

    def host_list(self, paging=None, **args) -> HostList:
        return HostList(self._api, paging=paging, **args)

    def host_prototype(self, **args) -> HostPrototype:
        return HostPrototype(self._api, **args)

    def host_prototype_list(self, paging=None, **args) -> HostPrototypeList:
        return HostPrototypeList(self._api, paging=paging, **args)

    def job(self, **args) -> Job:
        return Job(self._api, **args)

    def job_list(self, paging=None, **args) -> JobList:
        return JobList(self._api, paging=paging, **args)

    def prototype(self, **args) -> Prototype:
        return Prototype(self._api, **args)

    def prototype_list(self, paging=None, **args) -> PrototypeList:
        return PrototypeList(self._api, paging=paging, **args)

    def provider(self, **args) -> Provider:
        return Provider(self._api, **args)

    def provider_list(self, paging=None, **args) -> ProviderList:
        return ProviderList(self._api, paging=paging, **args)

    def provider_prototype(self, **args) -> ProviderPrototype:
        return ProviderPrototype(self._api, **args)

    def provider_prototype_list(self, paging=None, **args) -> ProviderPrototypeList:
        return ProviderPrototypeList(self._api, paging=paging, **args)

    def service(self, **args) -> Service:
        return Service(self._api, **args)

    def service_list(self, paging=None, **args) -> ServiceList:
        return ServiceList(self._api, paging=paging, **args)

    def service_prototype(self, **args) -> ServicePrototype:
        return ServicePrototype(self._api, **args)

    def service_prototype_list(self, paging=None, **args) -> ServicePrototypeList:
        return ServicePrototypeList(self._api, paging=paging, **args)

    def _upload(self, bundle_stream) -> Bundle:
        self._api.objects.stack.upload.create(file=bundle_stream)
        data = self._api.objects.stack.load.create(bundle_file="file")
        return self.bundle(bundle_id=data['id'])

    @allure_step('Upload bundle from "{1}"')
    def upload_from_fs(self, dirname, **args) -> Bundle:
        streams = stream.file(dirname, **args)
        if len(streams) > 1:
            raise TooManyArguments('upload_bundle_from_fs is not capable with building multiple \
                bundle editions from one dir. Use upload_from_fs_all instead.')
        return self._upload(streams[0])

    @allure_step('Upload bundles from "{1}"')
    def upload_from_fs_all(self, dirname, **args) -> BundleList:
        streams = stream.file(dirname, **args)
        # Create empty bundle list by filtering on wittingly nonexisting field
        # and value.
        result = BundleList(self._api, empty_bundlelist='not_existing_bundle')
        for st in streams:
            result.append(self._upload(st))
        return result

    @allure_step('Upload bundle from "{1}"')
    def upload_from_url(self, url) -> Bundle:
        return self._upload(stream.web(url))

    @allure_step("Delete bundle")
    def bundle_delete(self, **args):
        self._api.objects.stack.bundle.delete(bundle_id=self.bundle(**args).bundle_id)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import random
import socket
import docker
from adcm_client.util.wait import wait_for_url
from adcm_client.wrappers.api import ADCMApiWrapper


MINDOCKERPORT = 8000
MAXDOCKERPORT = 9000

DEFAULTIP = '127.0.0.1'


class UnableToBind(Exception):
    pass


def _port_is_free(ip, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((ip, port))
    if result == 0:
        return False
    return True


def _find_random_port(ip):
    for _ in range(0, 20):
        port = random.randint(MINDOCKERPORT, MAXDOCKERPORT)
        if _port_is_free(ip, port):
            return port
    raise UnableToBind("There is no free port for Docker after 20 tries.")


class ADCM():
    """
    Class that wraps ADCM Api operation over self.api (ADCMApiWrapper)
    and wraps docker over self.container (see docker module for info)
    """

    def __init__(self, container, ip, port):
        self.container = container
        self.ip = ip
        self.port = port
        self.url = 'http://{}:{}'.format(self.ip, self.port)
        self.api = ADCMApiWrapper(self.url)

    def stop(self):
        """Stops container"""
        self.container.stop()


class DockerWrapper():
    """Allow to connect to local docker daemon and spawn ADCM intances."""

    def __init__(self):
        self.client = docker.from_env()

    # pylint: disable=R0913
    def run_adcm(self, image='ci.arenadata.io/adcm',
                 remove=True, pull=True, name=None, tag=None, ip=DEFAULTIP, volumes=None):
        """
        Run ADCM in docker image.
        Return ADCM instance.

        Example:
        adcm = docker.run(image='ci.arenadata.io/adcm', tag=None, ip='127.0.0.1')

        If tag is None or is not present than it checks ADCM_TAG env
        variable and use it as image's tag. If there is no ADCM_TAG than
        it uses latest tag.
        """
        if tag is None:
            if "ADCM_TAG" in os.environ:
                tag = os.environ["ADCM_TAG"]
            else:
                tag = "latest"
        if pull:
            self.client.images.pull(image, tag)
        port = _find_random_port(ip)
        container = self.client.containers.run(
            "{}:{}".format(image, tag),
            ports={'8000': (ip, port)},
            volumes=volumes,
            remove=remove,
            detach=True,
            name=name
        )
        wait_for_url("http://{}:{}/api/v1/".format(ip, port), 60)
        return ADCM(container, ip, port)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=R0901
from adcm_client.util.search import search_one, search
from adcm_client.wrappers.api import ADCMApiWrapper
from collections import UserList, OrderedDict
from pprint import pprint
from time import sleep
import coreapi


def pp(*args, **kwargs):
    pprint("--------------------------------------------------")
    if args != []:
        pprint(args)
    if kwargs != {}:
        pprint(kwargs)
    pprint("--------------------------------------------------")


def strip_none_keys(args):
    """ Usefull function to interact with coreapi.

    While it's ok to pass function arguments with default equal to None,
    it is not allowed to pass it over coreapi. So we have to strip keys
    with None values.
    """
    return {k: v for k, v in args.items() if v is not None}


class TooManyArguments(Exception):
    pass


class NoSuchEndpoint(Exception):
    pass


class ObjectNotFound(Exception):
    pass


class ActionHasIssues(Exception):
    pass


class WaitTimeout(Exception):
    pass


class ADCMApiError(Exception):
    pass


class ResponseTooLong(Exception):
    """Response is too long, use paginated request"""


class PagingEnds(Exception):
    """There are no more data in paginated mode."""


class Paging:
    def __init__(self, paged_object, limit=50, **args):
        self._paged_object = paged_object
        self._limit = limit
        self._query_params = args      # Just passing it to paged_object
        self._offset = 0               # Offset in paging
        self._current_list = None      # Current page data
        self._current_iterator = None

    def __iter__(self):
        self._offset = 0
        return self

    def __next_list(self):
        try:
            result = self._paged_object(
                paging={
                    'limit': self._limit,
                    'offset': self._offset
                },
                **self._query_params
            )
            self._offset += self._limit
            self._current_list = result
            self._current_iterator = iter(self._current_list)
        except PagingEnds:
            raise StopIteration

    def __next_element(self):
        return next(self._current_iterator)

    def __next__(self):
        if self._current_iterator is None:
            self.__next_list()

        while True:
            try:
                result = next(self._current_iterator)
                return result
            except StopIteration:
                self.__next_list()


def _merge(*args, **kwargs):
    result = {}
    for i in args:
        result.update(i)
    result.update(kwargs)
    return result


def _find_endpoint(api, path):
    result = api
    if path is not None:
        for i in path:
            if i in result.__dict__:
                result = result.__dict__[i]
            else:
                raise NoSuchEndpoint
    return result


class EndPoint:
    def __init__(self, api, idname, path, path_args=None, awailable_filters=None):
        if idname is None:
            raise NotImplementedError

        self.point = _find_endpoint(api.objects, path)

        self.idname = idname

        if awailable_filters is None:
            self.awailable_filters = []
        elif isinstance(awailable_filters, list):
            self.awailable_filters = awailable_filters
        else:
            raise NotImplementedError

        if path_args is None:
            self.path_args = {}
        else:
            self.path_args = path_args

    def _get_filters_value(self, **args):
        result = {}
        for v in self.awailable_filters:
            # There is a dirty case when we have one Class for two endpoints
            # That is Host class. When it works as Cluster child (cluster/4/host)
            # it has cluster_id in path. But when it used as single object (/host/4)
            # it has cluster_id in filters
            if v in args and v not in self.path_args:
                result[v] = args[v]
        return result

    def list(self, paging=None, **args):
        if paging is None:
            paging = {}
        filters = self._get_filters_value(**args)
        try:
            result = self.point.list(**self.path_args, **paging, **filters)
        except coreapi.exceptions.ErrorMessage as e:
            # pylint: disable=W0212
            if "code" in e.error._data and e.error._data["code"] == "TOO_LONG":
                raise ResponseTooLong

        if isinstance(result, OrderedDict):
            # It's paging mode
            if result['results'] == []:
                raise PagingEnds
            return result['results']
        return result

    def read(self, object_id):
        return self.point.read(**self.get_object_path(object_id))

    def search(self, paging=None, **args):
        # TODO: Add filtering on backend
        return search(self.list(paging, **args), **args)

    def search_one(self, **args):
        # FIXME: paging
        if self.idname in args:
            return self.read(args[self.idname])
        data = search_one(self.list(**args), **args)
        if data is None:
            raise ObjectNotFound

        return self.read(data['id'])

    def get_object_path(self, object_id):
        return _merge(self.path_args, {self.idname: object_id})

    def get_subpoint(self, *path):
        return _find_endpoint(self.point, path)

    def delete(self, object_id):
        return self.point.delete(**self.get_object_path(object_id))


class BaseAPIObject:
    """That is common object for single ADCM's object"""
    IDNAME = None  # Will not be None in child
    PATH = None  # Will not be None in child
    FILTERS = []

    def _register_attrs(self):
        for k, v in self._data.items():
            if hasattr(self, k) and not callable(getattr(self, k)):
                setattr(self, k, v)

    def _copy_path_args(self, *names):
        for i in names:
            self._path_args[i] = self._data[i]

    def _merge(self, *args, **kwargs):
        return _merge(self._endpoint.get_object_path(self.id), *args, **kwargs)

    def __init__(self, api: ADCMApiWrapper, path=None, path_args=None, **args):
        if path is None and self.PATH is None and self.SUBPATH is None:
            raise NotImplementedError

        if path is None:
            # When we have some solid object like cluster,
            # it has a PATH list in class
            path = self.PATH

        if self.PATH is None:
            # If we have some object which is not hight level, like
            # cluster/<id>/action - host/<id>/action
            # In that case action is subobject and path should be
            # constructed somewhere out of scope.
            self.PATH = path

        self._endpoint = EndPoint(api, self.IDNAME, path, path_args, self.FILTERS)
        self._api = api

        self._client = api.objects

        self._data = self._endpoint.search_one(**args)

        if self._data is None:
            raise ObjectNotFound

        # TODO: Do something with id name consistency in Engine
        self._data[self.IDNAME] = self._data['id']
        # self._copy_path_args(self.IDNAME)
        self._register_attrs()

    def reread(self):
        self._data = self._endpoint.read(self.id)
        if self._data is None:
            raise ObjectNotFound
        self._register_attrs()

    def wait_for_attr(self, attrname, values, timeout=None, interval=None):
        if timeout is None:
            timeout = 86400
        if interval is None:
            interval = 0.2
        i = 0
        while getattr(self, attrname) not in values and i < timeout:
            sleep(interval)
            i = i + interval
            self.reread()
        if self.status in values:
            return self.status
        raise WaitTimeout

    def _subobject(self, classname, **args):
        return classname(self._api,
                         path=self.PATH + classname.SUBPATH,
                         path_args=self._endpoint.get_object_path(self.id),
                         **args)

    def _subcall(self, *path, **args):
        func = self._endpoint.get_subpoint(*path)
        return func(**self._merge(**args))

    def _child_obj(self, classname, **args):
        return classname(self._api, **_merge(args, {self.IDNAME: self.id}))

    def _parent_obj(self, classname):
        return classname(self._api, **{classname.IDNAME: self._data[classname.IDNAME]})

    def delete(self):
        return self._endpoint.delete(self.id)


class BaseAPIListObject(UserList):  # pylint: disable=too-many-ancestors
    """That is common object for multiple ADCM's object"""
    _ENTRY_CLASS = BaseAPIObject

    def __init__(self, api: ADCMApiWrapper, path=None, path_args=None, paging=None, **args):
        self._api = api
        self._client = api.objects
        if path_args is None:
            self._path_args = {}
        else:
            self._path_args = path_args

        if path is None:
            path = self._ENTRY_CLASS.PATH

        self._endpoint = EndPoint(api,
                                  self._ENTRY_CLASS.IDNAME,
                                  path,
                                  path_args,
                                  self._ENTRY_CLASS.FILTERS)
        data = []
        for i in self._endpoint.search(**args, paging=paging):
            data.append(self._ENTRY_CLASS(api,
                                          path=path,
                                          path_args=path_args,
                                          **{self._ENTRY_CLASS.IDNAME: i['id']}))
        super().__init__(data)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import os
import random
import string
from itertools import chain

import jinja2
import yaml
from docker import from_env
from docker.client import DockerClient
from docker.errors import ImageNotFound
from docker.models.containers import Container  # pylint: disable=unused-import
from docker.models.images import Image


class NoModulesToInstall(Exception):
    def __init__(self, message, errors=None):
        super().__init__(message)
        self.errors = errors


def _get_top_dirs(image: Image, prepared_image: Image, client: DockerClient) -> "list":
    """Returns a list of paths to all top level folders
     and files of python module that must be installed in image
    :param image: image without intaled modules
    :type image: Image
    :param prepared_image: image with intaled modules
    :type prepared_image: Image
    :param d_client: docker client
    :type d_client: DockerClient
    :return: list of path on image fs to module files
    :rtype: list
    """
    # list of packages from image without installed python modules
    default_module_list = _get_modules_list(image, client)

    # list of packages from image with installed installed modules
    modified_module_list = _get_modules_list(prepared_image, client)

    # list of packages tham must be installed
    modules = list(map(
        lambda x: x.split('==')[0],
        set(modified_module_list).difference(default_module_list)
    ))

    modules_data = yaml.safe_load_all(
        client.containers.run(
            prepared_image,
            'pip show -f %s' % ' '.join(modules),
            remove=True
        ).decode("utf-8")
    )
    return list(chain.from_iterable(map(
        lambda x: [os.path.join(x['Location'], i) for i in list(
            dict.fromkeys(
                map(
                    lambda y: os.path.normpath(y).split(os.sep)[0],
                    x['Files'].split()
                )
            )) if i not in ['..', '.']],
        modules_data
    )))


def _get_modules_list(image: Image, client: DockerClient) -> "list":
    """Run pip freeze in docker container from given image.
    Returns output as a list.

    :param image: image name
    :type image: Image
    :param client: docker client
    :type client: DockerClient
    :return: list of installed python pkgs in given inamge freeze format
    :rtype: list
    """
    return client.containers.run(
        image, '/bin/sh -c "pip freeze"', remove=True).decode("utf-8").split()


def _get_prepared_container(pkgs: list, image: Image, client: DockerClient) -> "Container":
    """Install python pkgs to container from given image

    :param pkgs: List of python packages
    :type pkgs: list
    :param image: image name
    :type image: Image
    :param client: docker client
    :type client: DockerClient
    :return: container with installed python packages
    :rtype: Container
    """
    command = '/bin/sh -c "'
    if pkgs.get('system_pkg'):
        command += 'apk add ' + ' '.join(pkgs.get('system_pkg')) + ' >/dev/null ;'
    command += ' pip install ' + ' '.join(pkgs.get('python_mod')) + ' >/dev/null"'
    return client.containers.run(image, command, detach=True)


def _copy_pkgs_files(path, dirs, image: Image, volumes: dict, client: DockerClient):
    """Copy list of dirs from container from given image to path.

    :param path: path where to copy
    :type path: str
    :param dirs: paths what need to be copied
    :type dirs: list
    :param image: image that contains needed paths
    :type image: Image
    :param volumes: volumes that must be mounted to container.
    :type volumes: dict
    :param client: docker client
    :type client: DockerClient
    """
    dirs = list(dict.fromkeys(dirs))  # filter on keys of duplicate elements
    command = '/bin/sh -c "mkdir %s/pmod; cp -r %s %s/pmod ;' % (path,
                                                                 ' '.join(dirs),
                                                                 path)
    command += ' chown -R %s %s/pmod"' % (os.getuid(), path)
    client.containers.run(image, command, volumes=volumes, remove=True)


def _get_prepared_image(pkgs, image: Image, client: DockerClient) -> "Image":
    """Install pgks to container from given image and commits container to a new image

    :return: returns image with installed required python packages
    :rtype: Image
    """
    container = _get_prepared_container(pkgs, image, client)
    container.wait()

    prepared_image_name = [
        image.tags[0].split(':')[0],
        ''.join(random.sample(string.ascii_lowercase, 5))
    ]
    prepared_image = container.commit(repository=prepared_image_name[0],
                                      tag=prepared_image_name[1])
    container.remove()
    return prepared_image


def python_mod_req(source_path, workspace, **kwargs):
    with open(os.path.join(source_path, kwargs['requirements']), 'r') as file:
        pkgs = yaml.safe_load(file)
        if pkgs.get('python_mod'):
            client = from_env()
            # choose image where to install python pkgs
            # by default adcm:latest but i think this may be bad practice
            # better to use adcm_min_version of bundle
            image_name = "arenadata/adcm:latest" if not kwargs.get('image') else kwargs['image']
            image = client.images.pull(image_name)
            # clean up flag
            rm_prepared_image = True

            # prepared image is an image with intalled python packages
            # that need to be packed in bundle
            # for debug purposes there may be an prepared_image variable received from spec.yaml
            if kwargs.get('prepared_image'):
                try:
                    prepared_image = client.images.get(kwargs['prepared_image'])
                    rm_prepared_image = False
                except ImageNotFound:
                    prepared_image = _get_prepared_image(pkgs, image, client)
            else:
                prepared_image = _get_prepared_image(pkgs, image, client)

            # list of all highlevel dirs of python pkgs that must be packed
            dirs = _get_top_dirs(image, prepared_image, client)

            # volume that contains workspace
            volumes = {
                workspace: {'bind': workspace, 'mode': 'rw'}
            }

            _copy_pkgs_files(source_path, dirs, prepared_image, volumes, client)

            if rm_prepared_image:
                client.images.remove(prepared_image.id)

        else:
            raise NoModulesToInstall('Can`t get python modules list to be inatalled')


def splitter(*args, **kwargs):
    env = jinja2.Environment(loader=jinja2.FileSystemLoader(args[0]),
                             undefined=jinja2.StrictUndefined)
    for file in kwargs['files']:
        tmpl = env.get_template(file)
        with codecs.open(os.path.join(args[0], (os.path.splitext(file)[0])), 'w', 'utf-8') as f:
            f.write(tmpl.render(kwargs['edition']))


def get_type_func(tpe):
    types = {'python_mod_req': python_mod_req, 'splitter': splitter}
    return types[tpe]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import sys
from fnmatch import fnmatch
from os import DirEntry, chdir, getcwd, scandir
from os.path import normpath


class NotValidSpecVersion(Exception):
    def __init__(self, message, errors=None):
        super().__init__(message)
        self.errors = errors


def compare_helper(version, except_list):
    if version is None:
        except_list.extend([
            'spec.yaml', 'pylintrc', '.[0-9a-zA-Z]*', '*pycache*',
            'README.md', '*requirements*', '*.gz', '*.md'])

        def v_None(sub: DirEntry):
            for n in except_list:
                if fnmatch(normpath(sub.path), n) or fnmatch(sub.name, n):
                    return True
            return False
        return v_None
    elif version == "1.0":
        prog = [re.compile(i) for i in except_list]

        def v_1_0(sub: DirEntry):
            for n in prog:
                if n.match(normpath(sub.path)):
                    return True
            return False
        return v_1_0
    else:
        raise NotValidSpecVersion('Not valid spec version').with_traceback(sys.exc_info()[2])


def add_to_tar(version, directory, except_list, tar):
    comparator = compare_helper(version, except_list)
    cwd = getcwd()
    chdir(directory)

    def _add_to_tar(path='./'):
        for sub in scandir(path):
            if not comparator(sub=sub):
                if sub.is_dir():
                    _add_to_tar(path=normpath(sub.path))
                else:
                    tar.add(normpath(sub.path))
    _add_to_tar()
    chdir(cwd)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import requests


def get_token(url="http://127.0.0.1:8000/api/v1/token", username="admin", password="admin"):
    result = requests.request("POST", url, data={"username": username, "password": password})
    # TODO: Error handling
    return result.json()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import socket
from time import sleep
from time import time as now
from urllib.parse import urlparse

import urllib3

# pylint: disable=W0703


def wait_net_service(server, port, timeout, period=1):
    """ Wait for network service to appear
        @param timeout: in seconds, if None or 0 wait forever
        @return: True of False
    """

    start = now()

    s = socket.socket()
    while now() - start < timeout:
        try:
            s.connect((server, port))
        except Exception:
            pass

        else:
            s.close()
            return True
        sleep(period)
    return False


def wait_for_url(url, timeout, period=1):
    """ Wait for url to responce something with http 200
        @param timeout: in seconds
        @param period: in seconds
        @return: True if connect successfull, False if not
    """
    # Disable annoying warning:
    # connectionpool.py          662 WARNING  Retrying
    logging.getLogger("urllib3").setLevel(logging.ERROR)

    urlparsed = urlparse(url)
    start = now()
    if not wait_net_service(urlparsed.hostname, urlparsed.port, timeout, period):
        return False

    http = urllib3.PoolManager()
    while now() - start < timeout:
        try:
            r = http.request('GET', url)
            if r.status == 200:
                return True
        except Exception:
            pass
        sleep(period)

    return False



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import gzip
import io
import os

import requests
from adcm_client.packer.bundle_build import build


def file(path, **args):
    if os.path.isdir(path):
        return list(build(repopath=path, **args).values())
    else:
        stream = io.BytesIO()
        try:
            stream.write(gzip.open(path, 'rb').read())
        except OSError:
            stream.write(io.open(path, 'rb').read())
        stream.seek(0)
        return [stream]


def web(url):
    response = requests.get(url=url)
    response.raise_for_status()
    return response.content



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import yaml
from .objects import ADCMClient


class ObjectNotFound(Exception):
    pass


class IncorrectOperation(Exception):
    pass


class Mapper:

    def __init__(self, filename, client: ADCMClient = None):
        self._filename = filename
        self._workdir = os.path.dirname(filename)
        os.chdir(self._workdir)
        self._client = client
        self._parse()
        self._vars = {}
        if client is not None:
            self._vars = {'external_adcm': client}

    def _parse(self):
        with open(self._filename, 'r') as stream:
            self._data = yaml.load(stream)

    def _do_rec(self, point):
        if point['object'] not in self._vars:
            raise ObjectNotFound
        object_var = self._vars[point['object']]
        if not hasattr(object_var, point['func']):
            raise IncorrectOperation
        function = getattr(object_var, point['func'])
        if not callable(function):
            raise IncorrectOperation
        self._vars[point['return']] = function(**point['args'])

    def do(self):
        for rec in self._data:
            self._do_rec(rec)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from time import gmtime, strftime

from git import Repo
from git.exc import InvalidGitRepositoryError

from .data.config_data import ConfigData


class NoVersionFound(Exception):
    def __init__(self, message, errors=None):
        super().__init__(message)
        self.errors = errors


class RestrictedSymbol(Exception):
    def __init__(self, message, errors=None):
        super().__init__(message)
        self.errors = errors


def add_build_id(path, reponame, edition, master_branches: list):
    def write_version(file, old_version, new_version):
        with open(file, 'r+') as config:
            data = config.read()
            data = data.replace(old_version, new_version)
            config.seek(0)
            config.truncate()
            config.write(data)

    edition = "community" if edition is None or edition == "None" else edition
    bundle = ConfigData(catalog=path)
    version = bundle.get_data('version', 'catalog', explict_raw=True)

    if version is None:
        raise NoVersionFound('No version detected').with_traceback(sys.exc_info()[2])

    try:
        git = Repo(path).git
        if git.describe('--all').split('/')[0] == 'tags':
            tag = git.describe('--all')
            branch = [out.split('/')[2] for out in git.branch('-a', '--contains', tag).splitlines()
                      if 'origin' in out][0]
        else:
            try:
                branch = git.describe('--all').split('/')[2]
            except IndexError:
                branch = git.rev_parse('--abbrev-ref', 'HEAD')
        if branch in master_branches:
            branch = '-1'
        elif git.describe('--all').split('/')[1] == 'pr':
            branch = '-rc' + branch + '.' + strftime("%Y%m%d%H%M%S", gmtime())
        else:
            branch = '-' + branch.replace("-", "_")
    except InvalidGitRepositoryError:
        branch = ''
    else:
        if not isinstance(version, str):
            raise TypeError('Bundle version must be string').with_traceback(sys.exc_info()[2])
        if '-' in version:
            raise RestrictedSymbol('Version contains restricted symbol \
                "-" in position %s' % version.index('-')).with_traceback(sys.exc_info()[2])
        write_version(bundle.file, version, version + branch)

    return str(reponame) + '_v' + str(version) + branch + '_' + edition + '.tgz'



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import sys
import tarfile
from distutils.dir_util import copy_tree, remove_tree
from io import BytesIO
from tempfile import mkdtemp

from .add_to_tar import add_to_tar
from .naming_rules import add_build_id
from .spec import SpecFile, spec_processing


def _prepare_ws(reponame, workspace, src_path, spec: SpecFile):
    edition_dirs = {}
    tmpdir = mkdtemp(prefix=reponame + '_', dir=workspace)
    for edition in spec.data['editions']:
        edition_dirs.update({edition['name']: os.path.join(tmpdir, str(edition['name']))})
        copy_tree(src_path, edition_dirs[edition['name']])
    return tmpdir, edition_dirs


def _prepare_result_dir(workspace, tarball_path):
    if tarball_path:
        os.makedirs(tarball_path, exist_ok=True)
        return tarball_path
    else:
        return workspace


def _pack(reponame, repopaths, tarpaths, spec: SpecFile, **kwargs):
    for edition in spec.data['editions']:
        name = edition.get('name')
        tarpath = tarpaths[name] if isinstance(tarpaths, dict) else tarpaths
        repopath = repopaths[name]
        tar_except = edition.get('exclude', [])

        # naming rules
        tarname = add_build_id(repopath, reponame, name, kwargs['master_branches'])

        stream = BytesIO()
        tar = tarfile.open(fileobj=stream, mode='w|gz')
        add_to_tar(spec.data['version'], repopath, tar_except, tar)
        logging.info("#######\n Edition %s \n#######", name)
        logging.info("#######\n Packed files list:\n%s\n#######", "\n".join(tar.getnames()))
        tar.close()
        stream.seek(0)
        # saving tarball
        yield os.path.join(tarpath, tarname), stream


def _clean_ws(path):
    if isinstance(path, dict):
        for i in path.values():
            remove_tree(i)
    else:
        remove_tree(path)


def build(reponame=None, repopath=None, workspace='/tmp',  # pylint: disable=R0913
          tarball_path=None, loglevel='ERROR',
          clean_ws=True, master_branches=None, **args):
    """Moves sources to workspace inside of temporary directory. \
    Some operations over sources cant be proceed concurent(for exemple in pytest with xdist \
    plugin) that why each thread need is own tmp dir with sources. \
    Also when there is complex docker containers launching to process some information there is \
    necessery to share same workspace with every used container.
    Proceed spec file.
    Writes build number to bundle config file.
    Recursively add files to bytes stream.
    Cleanup tmp dirs.

    :param repopath: Where bundle sources are
    :type repopath: str
    :param reponame: arenadata repository name. Used for naming aftifact and tmp dir.
    :type reponame: str
    :param workspace: where build operations will be performed, defaults to /tmp.
    :type workspace: str, optional
    :param tarball_path: where to copy builded bundle, defaults to None.
    None means that tarball will be left in temporary directory inside of workspace.
    :type tarball_path: str, optional
    :param loglevel: lower or equal to INFO will be stdout
    :type loglevel: str, optional
    :return: return a dict.
    Keys - path and name of tarball to save.
    Value - stream of bytes.
    :rtype: dict
    """
    if not master_branches:
        master_branches = ['master']
    if not repopath:
        raise ValueError('path to source should be defined')
    if not reponame:
        reponame = os.path.basename(os.path.realpath(repopath))

    logging.basicConfig(stream=sys.stdout, level=getattr(logging, loglevel))
    spec = SpecFile(os.path.join(repopath, 'spec.yaml'))
    spec.normalize_spec()

    ws_tepm_dir, work_dir_paths = _prepare_ws(reponame, workspace, repopath, spec)

    tarpath = _prepare_result_dir(workspace, tarball_path)
    spec_processing(spec, work_dir_paths, workspace)

    out = dict(
        _pack(
            reponame,
            work_dir_paths,
            tarpath,
            spec,
            master_branches=master_branches))

    if clean_ws:
        _clean_ws(ws_tepm_dir)

    return out



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=W0611, W0621, W0404, W0212, C1801
import coreapi
import logging
import requests

try:
    # pylint: disable=unused-import
    from pytest import allure  # noqa: F401
    # pylint: disable=unused-import
    import pytest
    IS_ALLURE = True
except ImportError:
    IS_ALLURE = False


logging.getLogger("urllib3").setLevel(logging.ERROR)


class APINode():
    pass


class ADCMApiError(Exception):
    pass


class ADCMApiWrapper():
    """Thin wrapper over ADCM API with coreapi (search django rest framework)
    Quick start:

    api =  ADCMApiWrapper()
    api.auth(username='admin', password='admin')

    Following function are equal and returns OrderedDict with API response:
    api.action(['cluster', 'list'])
    api.objects.cluster.list()

    Create and remove object:
    cluster = api.objects.cluster.create(name="test")
    api.objects.cluster.delete(cluster_id=cluster['id'])
    """

    def _fabric_function(self, node, path=None):
        if path is None:
            path = []

        def result(**kvargs):
            return self.action(path, params=kvargs)

        return result

    def _fabric_function_allure(self, node, path=None):
        # pylint: disable=no-member
        if path is None:
            path = []

        @pytest.allure.step(path[-1].title() + ' ' + path[-2])
        def result(**kvargs):
            return self.action(path, params=kvargs)

        return result

    def _parse_schema(self, node, is_allure=False, path=None):
        if path is None:
            path = []

        result = APINode()
        for funcname in node.links.keys():
            if is_allure:
                func = self._fabric_function_allure(node.links[funcname], path + [funcname])
            else:
                func = self._fabric_function(node.links[funcname], path + [funcname])
            setattr(result, funcname, func)

        for subnodename in node.data.keys():
            setattr(result,
                    subnodename,
                    self._parse_schema(node.data[subnodename], is_allure, path + [subnodename]))
        return result

    def __init__(self, url):
        """
        Init class with ADCM url.

        Example:
        api = ADCMApiWrapper('http://127.0.0.1:8000')
        """
        self.api_url = "/api/v1/"
        self.url = url
        self.client = None
        self.schema = None
        self.objects = None
        self.api_token = None

    def _check_for_error(self, data):
        if data is not None:
            if 'level' in data and data['level'] == 'error':
                raise ADCMApiError(data['code'], data['desc'])

    def auth(self, username, password):
        """Auth api client in ADCM and get schema"""

        result = requests.request(
            'POST', self.url + '/api/v1/token/',
            data={'username': username, 'password': password})
        token = result.json()
        self._check_for_error(token)
        auth = coreapi.auth.TokenAuthentication(
            scheme='Token',
            token=token['token']
        )
        self.client = coreapi.Client(auth=auth)
        self.schema = self.client.get("{}{}schema/".format(self.url, self.api_url))
        self.objects = self._parse_schema(self.schema, is_allure=IS_ALLURE)
        self.api_token = token['token']

    def action(self, *args, **kvargs):
        """
        Do operation over api. For information see coreapi documentation.

        Example:
        api.action(['cluster', 'create'], name='testcluster')
        """
        data = self.client.action(self.schema, *args, **kvargs)
        self._check_for_error(data)
        return data



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import fnmatch
import glob
import io
import re
import tarfile

import requests
import yaml
from git import Git


# TO DO: Exeptions - non valid methods calls when required property is None
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
class ConfigData:
    def __init__(self, git: Git = None, file=None, data=None, catalog=None,
                 branch=None, tar=None, url=None):
        self.url = url
        self.tar = tar
        self.file = file
        self.data = data
        self.branch = branch
        self.remote_file = None
        self.catalog = catalog
        self.git = git
        self.switch = {
            'data': self._from_data,
            'url': self._from_url,
            'tar': self._from_tar,
            'file': self._from_file,
            'catalog': self._from_fs,
            'branch': self._from_remote_git
        }

    def get_data(self, key, provider, **kwargs):
        """Interface to get value of requested key from bundle config

        :param key: request key
        :type key: str
        :param provider: supported providers: data(dict), url, tar, file, catalog, branch
        :type provider: str
        :return: requsted keys
        """
        return self.switch[provider](key, **kwargs)

    def _from_file(self, key, **kwargs):
        with open(self.file, 'r') as file:
            self.data = yaml.safe_load(file)
        return self._from_data(key)

    def _from_data(self, key, **kwargs):
        try:
            value = [entry.get(key) for entry in self.data
                     if (entry.get('type') == 'cluster'
                         or entry.get('type') == 'provider'  # noqa: W503
                         or entry.get('type') == 'host')]  # noqa: W503
        except (TypeError, AttributeError):
            value = [entry[1].get(key) for entry in self.data.items()
                     if (entry[0] == 'cluster' or entry[0] == 'host' or entry[0] == 'provider')]
        for idx, _ in enumerate(value):
            if value[idx]:
                value[0] = value[idx]
        return value[0]

    # get bundle version
    def _from_fs(self, key, **kwargs):
        # find config file

        regex = "/**/config.y*ml" if kwargs.get('explict_raw', False) else "/**/*config.y*ml"
        configs = glob.glob(str(self.catalog) + regex, recursive=True)

        # version detection
        value = None
        for conf in configs:
            try:
                self.file = conf
                value = self._from_file(key)
                break
            except IndexError:
                pass

        return value

    def _from_remote_git(self, key, **kwargs):
        # using ./*config* patern dont work, func is usig metasymbol as is
        configs = [config for config in
                   self.git.ls_tree('-r', '--name-only', self.branch).splitlines()
                   if re.match('.*config.y.*ml', config)]
        value = None
        for conf in configs:
            try:
                self.remote_file = conf
                self.data = yaml.safe_load(self.git.show("%s:%s" % (self.branch, conf)))
                value = self._from_data(key)
            except IndexError:
                pass
        return value

    def _from_tar(self, key, **kwargs):
        tar = tarfile.open(self.tar) if isinstance(self.tar, str)\
            else tarfile.open(fileobj=self.tar)

        def conf_files(members):
            for tarinfo in members:
                if fnmatch.fnmatch(tarinfo.name, '*config.y*ml'):
                    yield tarinfo

        confs = conf_files(tar)
        value = None

        for conf in confs:
            try:
                self.file = conf
                self.data = yaml.safe_load(tar.extractfile(conf).read().decode('utf-8'))
                value = self._from_data(key)
                break
            except IndexError:
                pass
        return value

    def _from_url(self, key, **kwargs):
        response = requests.get(url=self.url)
        response.raise_for_status()
        self.tar = io.BytesIO(response.content)
        return self._from_tar(key)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import sys
from os.path import join
from subprocess import check_output

import yaml

from .types import get_type_func


class SpecFile:

    def __init__(self, spec):
        try:
            with open(spec, 'r') as file:
                self.data = yaml.safe_load(file)
        except FileNotFoundError:
            self.data = {}

    def to_1_0(self):
        new_spec = dict([('version', None),
                         ('editions', [{
                             'name': None,
                             'exclude': self.except_var(self.data),
                             'preprocessors': []}])])
        for i in self.data.get('processing', {}):
            if i.get('script'):
                new_spec['editions'][0]['preprocessors'].append(
                    {
                        'type': 'script',
                        'script': join(self.data[i['name'] + '_dir'], i['script']),
                        'args': [i['file']]
                    }
                )
            elif i.get('name') == 'python_mod_req':
                new_spec['editions'][0]['preprocessors'].append(
                    {
                        'type': i['name'],
                        'requirements': i['file']
                    }
                )
            else:
                sys.exit('Used unrecognized func:%s' % i.get('name'))
        return new_spec

    def normalize_spec(self):
        versions = ['1.0']
        migrations = dict([('1.0', self.to_1_0)])
        index = versions.index(self.data.get('version')) + 1\
            if self.data.get('version') in versions else 0
        for i in versions[index:]:
            self.data = migrations[i]()

        return self.data

    # deprecated method. Needed for backward compatibility with old specs
    def except_var(self, config):
        tar_except = []
        for k, v in config.items():
            if '_dir' in k:
                tar_except.append(v)
        for k in config.get('processing', {}):
            if k.get('except_file', False):
                tar_except.append(k.get('file'))
        return tar_except


def spec_processing(spec: SpecFile, path, workspace):
    for edition in spec.data['editions']:
        if edition.get('preprocessors'):
            for x in edition['preprocessors']:
                if x.get('script'):
                    command = [x['script']]
                    if x.get('args'):
                        command.extend(x['args'])
                    logging.info(check_output(command, cwd=path[edition['name']]).decode("utf-8"))
                else:
                    jinja_values = {'edition': edition['name']}
                    get_type_func(x['type'])(path[edition['name']], workspace,
                                             edition=jinja_values, **x)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import setuptools

test_deps = [
    'pytest',
]
extras = {
    'test': test_deps,
}

setuptools.setup(
    name="adcm_client",
    version="2020.02.25.17",
    author="Anton Chevychalov",
    author_email="cab@arenadata.io",
    description="ArenaData Cluster Manager Client",
    url="https://github.com/arenadata/adcm",
    packages=setuptools.find_packages(),
    install_requires=[
        'pyyaml', 'coreapi', 'ipython', 'gitpython', 'docker', 'jinja2',
        'version_utils'
    ],
    tests_require=test_deps,
    extras_require=extras,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    scripts=['bin/adcm_sdk_shell', 'bin/adcm_sdk_pack'],
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest


@pytest.fixture(scope="module")
def data():
    return [
        {
            "id": 1,
            "name": "ADB",
            "version": "5",
            "hash": "56daa2c4dccc2859ddd005ceed6b1f04a52a4a07",
            "description": "Arenadata Database",
            "date": "2018-08-20T13:24:37.649437Z",
            "url": "http://127.0.0.1:8000/api/v1/stack/bundle/1/",
            "update": "http://127.0.0.1:8000/api/v1/stack/bundle/1/update/"
        },
        {
            "id": 2,
            "name": "SSH Common",
            "version": "1.0",
            "hash": "f9f21250ebcadaaf9f1fd8a6b303e632d4159a75",
            "description": "",
            "date": "2018-08-20T13:25:19.090129Z",
            "url": "http://127.0.0.1:8000/api/v1/stack/bundle/2/",
            "update": "http://127.0.0.1:8000/api/v1/stack/bundle/2/update/"
        },
        {
            "id": 3,
            "name": "Digital Energy virtual machine",
            "version": "1.5",
            "hash": "865addbf0122f36a4de58aa58e0edb4a729528dc",
            "description": "",
            "date": "2018-08-20T13:26:35.018593Z",
            "url": "http://127.0.0.1:8000/api/v1/stack/bundle/3/",
            "update": "http://127.0.0.1:8000/api/v1/stack/bundle/3/update/"
        },
        {
            "id": 4,
            "name": "Monitoring",
            "version": "1.0",
            "hash": "2b64a9074dc7a869b433ecb40d4aeefc9d884800",
            "description": "Monitoring and Control Software",
            "date": "2018-08-20T13:26:41.564955Z",
            "url": "http://127.0.0.1:8000/api/v1/stack/bundle/4/",
            "update": "http://127.0.0.1:8000/api/v1/stack/bundle/4/update/"
        }
    ]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from .search import search, search_one, TooManyResult


def test_search_one_too_many_result(data):
    with pytest.raises(TooManyResult):
        result = search_one(data)
        assert result is None


def test_search_one_not_found(data):
    result = search_one(data, name='azaza')
    assert result is None


def test_search_one(data):
    result = search_one(data, id=1)
    assert result['name'] == 'ADB'


def test_search_found(data):
    result = list(search(data))
    assert len(result) == 4
    result = list(search(data, version="1.0"))
    assert len(result) == 2
    assert result[0]['id'] == 2 or result[1]['id'] == 2


def test_search_generator(data):
    for i in search(data):
        assert "date" in i


def test_search_not_found(data):
    # pylint: disable=unused-variable
    for i in search(data, description=1):
        assert False  # We just should not be there
    assert True
